/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 20/08/22.
//

#ifndef FGGL_ENTITY_GRIDWORLD_ZONE_HPP
#define FGGL_ENTITY_GRIDWORLD_ZONE_HPP

#include <array>
#include <vector>

#include "fggl/math/types.hpp"

namespace fggl::entity::grid {

	template<typename T, uint32_t width, uint32_t height>
	struct Grid {
		public:
			Grid() = default;

			inline T& at(math::vec2i pos) {
				return m_cells[getCellIndex(pos)];
			}

			const T& at(math::vec2i pos) const {
				return m_cells[getCellIndex(pos)];
			}

			inline bool inBounds(math::vec2i pos) const {
				return 0 <= pos.x && pos.x <= size.x &&
					   0 <= pos.y && pos.y <= size.y;
			}

		private:
			constexpr static math::vec2i size = math::vec2i(width, height);
			std::array<T, size.x * size.y> m_cells;


			inline uint32_t getCellIndex(math::vec2i pos) const {
				assert( inBounds(pos));
				return pos.y * size.x + pos.x;
			}
	};

	struct FloorTile {
		constexpr static uint8_t IMPOSSIBLE = 0;
		uint8_t moveCost = IMPOSSIBLE;
	};

	struct TileSet {
		std::vector<FloorTile> m_floors;
	};

	/**
	 * A 2D representation of a space.
	 *
	 * @tparam width the grid width in units
	 * @tparam height the grid height in units
	 */
	template<uint32_t width, uint32_t height>
	struct Area2D {
		public:
			inline explicit Area2D(TileSet& tiles) : m_tiles(tiles) {}

			inline FloorTile& floorAt(uint32_t x, uint32_t y) {
				return m_tiles.m_floors[ m_floors.at({x, y}) ];
			}

			inline bool canMove(math::vec2i pos) const {
				return m_tiles.m_floors[m_floors.getCell(pos)] != 0;
			}

			inline bool canMove(math::vec2i pos, math::vec2i dir) const {
				return canMove(pos + dir) && !blocked(pos, dir);
			}

			inline bool blocked(math::vec2i pos, math::vec2i dir) const {
				if ( dir.x == -1 || dir.y == -1 ) {
					return m_walls.getCell(pos) != 0;
				} else {
					m_walls.getCell(pos + dir) != 0;
				}
			}

			void neighbours(math::vec2i pos, std::vector<math::vec2i> &neighbours) const;
		private:
			TileSet& m_tiles;
			Grid<uint32_t, width, height> m_floors;
			Grid<uint32_t, width + 1, height + 1> m_walls;
	};

} // namespace fggl::entity::gridworld

#endif //FGGL_ENTITY_GRIDWORLD_ZONE_HPP
